@@ -5,6 +5,8 @@ References:
 PROV-DM: http://www.w3.org/TR/prov-dm/
 PROV-JSON: https://provenance.ecs.soton.ac.uk/prov-json/
 """
+from __future__ import (absolute_import, division, print_function,
+                        unicode_literals)
 
 __author__ = 'Trung Dong Huynh'
 __email__ = 'trungdong@donggiang.com'
@@ -21,12 +23,12 @@ from copy import deepcopy
 import io
 from prov import Error, serializers
 
-from StringIO import StringIO
-
 import os
 import shutil
 import tempfile
-from urlparse import urlparse
+
+import six
+from six.moves.urllib.parse import urlparse
 
 from prov.identifier import Identifier, QualifiedName, XSDQName
 from prov.constants import *
@@ -34,7 +36,7 @@ from prov.constants import *
 
 # Data Types
 def _ensure_datetime(value):
-    if isinstance(value, basestring):
+    if isinstance(value, six.string_types):
         return dateutil.parser.parse(value)
     else:
         return value
@@ -62,9 +64,10 @@ DATATYPE_PARSERS = {
 
 # Mappings for XSD datatypes to Python standard types
 XSD_DATATYPE_PARSERS = {
-    XSD_STRING: unicode,
+    XSD_STRING: six.text_type,
     XSD_DOUBLE: float,
-    XSD_LONG: long,
+    # long on Python 2, int on Python 3
+    XSD_LONG: six.integer_types[-1],
     XSD_INT: int,
     XSD_BOOLEAN: parse_boolean,
     XSD_DATETIME: parse_xsd_datetime,
@@ -80,12 +83,12 @@ first = lambda a_set: next(iter(a_set), None)
 
 
 def _ensure_multiline_string_triple_quoted(s):
-    format_str = u'"""%s"""' if isinstance(s, basestring) and '\n' in s else u'"%s"'
+    format_str = u'"""%s"""' if isinstance(s, six.string_types) and '\n' in s else u'"%s"'
     return format_str % s
 
 
 def encoding_provn_value(value):
-    if isinstance(value, basestring):
+    if isinstance(value, six.string_types):
         return _ensure_multiline_string_triple_quoted(value)
     elif isinstance(value, datetime.datetime):
         return u'"{0}" %% xsd:dateTime'.format(value.isoformat())
@@ -95,12 +98,13 @@ def encoding_provn_value(value):
         return u'"%i" %%%% xsd:boolean' % value
     else:
         # TODO: QName export
-        return unicode(value)
+        return six.text_type(value)
 
 
+@six.python_2_unicode_compatible
 class Literal(object):
     def __init__(self, value, datatype=None, langtag=None):
-        self._value = unicode(value)  # value is always a string
+        self._value = six.text_type(value)  # value is always a string
         if langtag:
             if datatype is None:
                 logger.debug('Assuming prov:InternationalizedString as the type of "%s"@%s' % (value, langtag))
@@ -115,13 +119,10 @@ class Literal(object):
                 )
                 datatype = PROV["InternationalizedString"]
         self._datatype = datatype
-        self._langtag = unicode(langtag) if langtag is not None else None  # langtag is always a string
-
-    def __unicode__(self):
-        return self.provn_representation()
+        self._langtag = six.text_type(langtag) if langtag is not None else None # langtag is always a string
 
     def __str__(self):
-        return unicode(self).encode('utf-8')
+        return self.provn_representation()
 
     def __repr__(self):
         return u'<Literal: %s>' % self.provn_representation()
@@ -151,9 +152,9 @@ class Literal(object):
     def provn_representation(self):
         if self._langtag:
             #  a language tag can only go with prov:InternationalizedString
-            return u'%s@%s' % (_ensure_multiline_string_triple_quoted(self._value), unicode(self._langtag))
+            return u'%s@%s' % (_ensure_multiline_string_triple_quoted(self._value), six.text_type(self._langtag))
         else:
-            return u'%s %%%% %s' % (_ensure_multiline_string_triple_quoted(self._value), unicode(self._datatype))
+            return u'%s %%%% %s' % (_ensure_multiline_string_triple_quoted(self._value), six.text_type(self._datatype))
 
 
 # Exceptions and warnings
@@ -167,15 +168,17 @@ class ProvWarning(Warning):
     pass
 
 
+@six.python_2_unicode_compatible
 class ProvExceptionInvalidQualifiedName(ProvException):
     def __init__(self, qname):
         self.qname = qname
 
-    def __unicode__(self):
+    def __str__(self):
         return u'Invalid Qualified Name: %s' % self.qname
 
 
 #  PROV records
+@six.python_2_unicode_compatible
 class ProvRecord(object):
     """Base class for PROV records."""
     FORMAL_ATTRIBUTES = ()
@@ -257,7 +260,7 @@ class ProvRecord(object):
             literal = literal.identifier
 
         if isinstance(literal, str):
-            return unicode(literal)
+            return six.text_type(literal)
         elif isinstance(literal, QualifiedName):
             return self._bundle.valid_qualified_name(literal)
         elif isinstance(literal, Literal) and literal.has_no_langtag():
@@ -327,11 +330,8 @@ class ProvRecord(object):
 
         return set(self.attributes) == set(other.attributes)
 
-    def __unicode__(self):
-        return self.get_provn()
-
     def __str__(self):
-        return unicode(self).encode('utf-8')
+        return self.get_provn()
 
     def get_provn(self):
         items = []
@@ -339,7 +339,7 @@ class ProvRecord(object):
         # Generating identifier
         relation_id = u''  # default blank
         if self._identifier:
-            identifier = unicode(self._identifier)  # TODO: QName export
+            identifier = six.text_type(self._identifier)  # TODO: QName export
             if self.is_element():
                 items.append(identifier)
             else:
@@ -351,7 +351,7 @@ class ProvRecord(object):
             if attr in self._attributes and self._attributes[attr]:
                 value = first(self._attributes[attr])  # Formal attributes always have single values
                 # TODO: QName export
-                items.append(value.isoformat() if isinstance(value, datetime.datetime) else unicode(value))
+                items.append(value.isoformat() if isinstance(value, datetime.datetime) else six.text_type(value))
             else:
                 items.append(u'-')
 
@@ -366,7 +366,7 @@ class ProvRecord(object):
                     except AttributeError:
                         provn_represenation = encoding_provn_value(value)
                     # TODO: QName export
-                    extra.append(u'%s=%s' % (unicode(attr), provn_represenation))
+                    extra.append(u'%s=%s' % (six.text_type(attr), provn_represenation))
 
         if extra:
             items.append(u'[%s]' % u', '.join(extra))
@@ -746,11 +746,11 @@ class NamespaceManager(dict):
             return XSDQName(new_qname) if is_xsd_qname else new_qname
 
         # Trying to guess from here
-        if not isinstance(qname, (basestring, Identifier)):
+        if not isinstance(qname, (six.string_types, Identifier)):
             # Only proceed for string or URI values
             return None
         # Try to generate a Qualified Name
-        str_value = qname.uri if isinstance(qname, Identifier) else unicode(qname)
+        str_value = qname.uri if isinstance(qname, Identifier) else six.text_type(qname)
         if str_value.startswith('_:'):
             # this is a blank node ID
             return None
@@ -790,7 +790,7 @@ class NamespaceManager(dict):
             return original_prefix
         count = 1
         while True:
-            new_prefix = '_'.join((original_prefix, unicode(count)))
+            new_prefix = '_'.join((original_prefix, six.text_type(count)))
             if new_prefix in self:
                 count += 1
             else:
@@ -928,10 +928,12 @@ class ProvBundle(object):
                     found = True
                     break
             if not found:
-                logger.debug("Equality (ProvBundle): Could not find this record: %s", unicode(record_a))
+                logger.debug("Equality (ProvBundle): Could not find this record: %s", six.text_type(record_a))
                 return False
         return True
 
+    __hash__ = None
+
     # Transformations
     def _unified_records(self):
         """Returns a list of unified records
@@ -1306,7 +1308,7 @@ class ProvDocument(ProvBundle):
 
     @property
     def bundles(self):
-        return frozenset(self._bundles.values())
+        return self._bundles.values()
 
     # Transformations
     def flattened(self):
@@ -1407,7 +1409,7 @@ class ProvDocument(ProvBundle):
         """
         serializer = serializers.get(format)(self)
         if destination is None:
-            stream = StringIO()
+            stream = io.StringIO()
             serializer.serialize(stream, **args)
             return stream.getvalue()
         if hasattr(destination, "write"):
@@ -1439,7 +1441,8 @@ class ProvDocument(ProvBundle):
         serializer = serializers.get(format)()
 
         if content is not None:
-            stream = StringIO(content)
+            # io.StringIO only accepts unicode strings (i.e. six.text_type)
+            stream = io.StringIO(content if isinstance(content, six.text_type) else six.u(content))
             return serializer.deserialize(stream, **args)
 
         if source is not None:
@@ -1471,7 +1474,7 @@ def sorted_attributes(element, attributes):
     # first and then sorting by the text, also including the namespace
     # prefix if given.
     sort_fct = lambda x: (
-        unicode(x[0]), unicode(x[1].value if hasattr(x[1], "value") else x[1]))
+        six.text_type(x[0]), six.text_type(x[1].value if hasattr(x[1], "value") else x[1]))
 
     sorted_elements = []
     for item in order:
