@@ -1,3 +1,6 @@
+from __future__ import (absolute_import, division, print_function,
+                        unicode_literals)
+
 from prov.model import ProvDocument
 from prov.tests.utility import BaseTestCase, RoundTripTestCase
 from prov.tests.test_model import AllTestsBase
